use crate::{auth::Backend, models::rating::Rating, page, suc, AppState, Station};
use axum::{
    extract::State,
    response::{IntoResponse, Redirect},
    routing::get,
    Router,
};
use axum_login::login_required;
use maud::{html, Markup};
use rand::{
    distr::{Distribution, Uniform},
    rng,
};
use route::Route;
use sqlx::SqlitePool;
use std::sync::Arc;
use team::Team;
use tower_sessions::Session;

pub(crate) mod route;
pub(crate) mod station;
pub(crate) mod team;
pub(crate) mod user;

fn generate_random_alphanumeric(length: usize) -> String {
    let mut rng = rng();
    let chars: Vec<char> = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ".chars().collect();
    let dist = Uniform::new(0, chars.len()).unwrap();

    (0..length).map(|_| chars[dist.sample(&mut rng)]).collect()
}

async fn highscore(State(db): State<Arc<SqlitePool>>, session: Session) -> Markup {
    let routes = Route::all(&db).await;

    let content = html! {
        h1 {
            a href="/admin" { "↩️" }
            (t!("highscore"))
        }
        @for (idx, route) in routes.into_iter().enumerate() {
            details open[idx==0] {
                summary { (route.name) }

                div class="overflow-auto" {
                    table {
                        thead {
                            tr {
                                td { (t!("team")) }
                                @for station in route.stations(&db).await {
                                    td {
                                        (station)
                                    }
                                }
                                td { (t!("total_points")) }
                                td { (t!("rank")) }
                                td { (t!("team")) }
                            }
                        }
                        tbody {
                            @let mut rank = 0;
                            @let mut amount_teams_iterated = 0;
                            @let mut prev_points = i64::MAX;
                            @for team in route.teams_ordered_by_points(&db).await {
                                @let mut total_points = 0;
                                ({ amount_teams_iterated += 1;"" })
                                tr {
                                    td {
                                        a href=(format!("/admin/team/{}", team.id)) {
                                            (team.name)
                                        }
                                    }
                                    @for station in route.stations(&db).await {
                                        td {
                                            @if let Some(rating) = Rating::find_by_team_and_station(&db, &team, &station).await {
                                                @if let (Some(notes), Some(points)) = (rating.notes, rating.points) {
                                                    ({total_points += points;""})
                                                    em data-placement="bottom" data-tooltip=(notes) { (points) }
                                                }@else if let Some(points) = rating.points {
                                                    ({total_points += points;""})
                                                    (points)
                                                }@else {
                                                    em data-placement="bottom" data-tooltip=(t!("station_has_not_rated_team_yet")) {
                                                        "?"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    td { (total_points) }
                                    td {
                                        @if total_points < prev_points {
                                            ({rank = amount_teams_iterated; ""})
                                            ({ prev_points = total_points;"" })
                                        }
                                        (rank)
                                        "."
                                    }
                                    td {
                                        a href=(format!("/admin/team/{}", team.id)) {
                                            (team.name)
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    };

    page(content, session, false).await
}

#[derive(PartialEq)]
pub enum RunStatus {
    NoStationsYet,
    Active,
    HasEnded,
}

impl RunStatus {
    pub async fn curr(db: &SqlitePool) -> Self {
        let stations = Station::all(db).await;
        if stations.is_empty() {
            return RunStatus::NoStationsYet;
        }

        if station::some_team_has_last_station_id(db).await {
            return RunStatus::HasEnded;
        }
        RunStatus::Active
    }
}

async fn index(State(db): State<Arc<SqlitePool>>, session: Session) -> Markup {
    let status = RunStatus::curr(&db).await;
    let content = html! {
        nav {
            ul {
                img class="logo" src="/logo-hor.svg" style="max-width: 100%; width: 25em; margin:auto;";
                img class="logo-inv" src="/logo-hor-inv.svg" width="max-width: 100%; width: 25em; margin:auto;";
            }
            ul {
                a href="/auth/logout" {
                    (t!("logout"))
                }
            }
        }
        nav {
            ul {
                li {
                    a role="button" href="/admin/station" {
                        (t!("stations"))
                    }
                }
                li {
                    a role="button" href="/admin/route" {
                        (t!("routes"))
                    }
                }
                li {
                    a role="button" href="/admin/team" {
                        (t!("teams"))
                    }
                }
                li {
                    a role="button" href="/admin/highscore" {
                        (t!("highscore"))
                    }
                }
                li {
                    a href="/admin/user" {
                        (t!("admins"))
                    }
                }
            }
        }
        @match status {
            RunStatus::NoStationsYet => {
                (t!("no_stations_yet"))
                (t!("change_that_below"))
                a role="button" href="/admin/station" {
                    (t!("go_to_stations"))
                }
            },
            RunStatus::Active  => {
                a href="/admin/end-run" onclick=(format!("return confirm('{}');", t!("confirm_end_run"))) {
                    button style="background-color: red;" {
                        (t!("end_run"))
                    }
                }
            },
            RunStatus::HasEnded => {
                @let stations = Station::all(&db).await;
                a href="/admin/restart-run" onclick=(format!("return confirm('{}');", t!("confirm_restart_run"))) {
                    button style="background-color: red;" {
                        (t!("restart_run"))
                    }
                }
                table {
                    thead {
                        tr {
                            th { (t!("stations")) }
                            th { (t!("come_home_with_these_groups")) }
                        }
                    }
                    tbody {
                        @for station in stations {
                            tr {
                                td { (station) }
                                td {
                                    ol {
                                        @for team in Team::all_with_last_station(&db, &station).await {
                                            li { (team) }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    };
    page(content, session, false).await
}

async fn end_run(State(db): State<Arc<SqlitePool>>, session: Session) -> impl IntoResponse {
    Team::end_run(&db).await;
    suc!(session, t!("run_ended"));
    Redirect::to("/admin")
}
async fn restart_run(State(db): State<Arc<SqlitePool>>, session: Session) -> impl IntoResponse {
    Team::restart_run(&db).await;
    suc!(session, t!("run_restarted"));
    Redirect::to("/admin")
}

pub(super) fn routes() -> Router<AppState> {
    Router::new()
        .route("/", get(index))
        .route("/highscore", get(highscore))
        .route("/end-run", get(end_run))
        .route("/restart-run", get(restart_run))
        .nest("/station", station::routes())
        .nest("/route", route::routes())
        .nest("/team", team::routes())
        .nest("/user", user::routes())
        .layer(login_required!(Backend, login_url = "/auth/login"))
}
