use super::Route;
use crate::{admin::station::Station, er, page, suc, AppState};
use axum::{
    extract::State,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Form, Router,
};
use maud::{html, Markup, PreEscaped};
use serde::Deserialize;
use sqlx::SqlitePool;
use std::sync::Arc;
use tower_sessions::Session;

async fn index(State(db): State<Arc<SqlitePool>>, session: Session) -> Markup {
    let routes = Route::all(&db).await;

    let content = html! {
        h1 {
            a href="/admin" { "↩️" }
            (t!("routes"))
        }
        article {
            em { (t!("routes")) }
            (PreEscaped(t!("routes_expl_without_first_word")))
        }
        ol {
            @for route in &routes{
                li {
                    @if route.stations(&db).await.is_empty() {
                        em data-tooltip=(t!("route_has_no_station_assigned")) {
                            "⚠️"
                        }
                    }
                     a href=(format!("/admin/route/{}", route.id)){
                        (route.name)
                     }
                     a href=(format!("/admin/route/{}/delete", route.id))
                       onclick=(format!("return confirm('{}');", t!("route_confirm_deletion"))) {
                         "🗑️"
                     }
                }
            }
        }
        @if routes.is_empty() {
            article class="warning" {
                (t!("no_routes"))
                " "
                (t!("change_that_below"))
            }
        }
        h2 { (t!("route_new")) }
        form action="/admin/route" method="post" {
            fieldset role="group" {
                input type="text" name="name" placeholder=(t!("route_name")) required;
                input type="submit" value=(t!("route_new"));
            }
        }
    };
    page(content, session, false).await
}

#[derive(Deserialize)]
struct CreateForm {
    name: String,
}

async fn create(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    Form(form): Form<CreateForm>,
) -> impl IntoResponse {
    match Route::create(&db, &form.name).await {
        Ok(()) => suc!(session, t!("route_create_succ", name = form.name)),
        Err(e) => er!(
            session,
            t!("route_create_err_duplicate_name", name = form.name, err = e)
        ),
    }

    Redirect::to("/admin/route")
}

async fn delete(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, id).await else {
        er!(session, t!("route_delete_err_nonexisting", id = id));
        return Redirect::to("/admin/route");
    };

    match route.delete(&db).await {
        Ok(()) => suc!(session, t!("route_delete_succ", name = route.name)),
        Err(e) => er!(
            session,
            t!("route_delete_err_already_used", name = route.name, err = e)
        ),
    }

    Redirect::to("/admin/route")
}

async fn view(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> Result<Markup, impl IntoResponse> {
    let Some(route) = Route::find_by_id(&db, id).await else {
        er!(session, t!("route_delete_err_nonexisting", id = id));
        return Err(Redirect::to("/admin/route"));
    };

    let cur_stations = route.stations(&db).await;
    let stations_not_in_route = route.stations_not_in_route(&db).await;

    let teams = route.teams(&db).await;

    let content = html! {
        h1 {
            a href="/admin/route" { "↩️" }
            (t!("route"))
             " "
            (route.name)
        }
        article {
            details {
                summary { (t!("route_name_edit")) "✏️" }
                form action=(format!("/admin/route/{}/name", route.id)) method="post" {
                    input type="text" name="name" value=(route.name) required;
                    input type="submit" value=(t!("save"));
                }
            }
        }
        h2 { (t!("stations")) }
        @if cur_stations.is_empty() {
            @if stations_not_in_route.is_empty() {
                article class="error" {
                    (t!("cant_assign_station_to_route_because_no_stations_exist"))
                    (PreEscaped(" &rarr; "))
                    a role="button" href="/admin/station" {
                        (t!("station_create"))
                    }
                }
            } @else {
                article class="warning" {
                    (t!("route_has_no_stations_yet"))
                    " "
                    (t!("change_that_below"))
                }
            }
        } @else {
            ol {
                @for (idx, station) in cur_stations.into_iter().enumerate() {
                    li {
                        (station)
                         @if idx > 0 {
                            a href=(format!("/admin/route/{}/move-station-higher/{}", route.id, station.id)){
                                em data-tooltip=(t!("station_move_up", name = station.name)) {
                                    "⬆️"
                                }
                            }
                         }
                        a href=(format!("/admin/route/{}/delete-station/{}", route.id, station.id))
                          onclick=(format!("return confirm('{}');", t!("confirm_remove_station_from_route"))) {
                            "🗑️"
                        }
                    }
                }
            }
        }
        @if !stations_not_in_route.is_empty(){
            form action=(format!("/admin/route/{}/add-station", route.id)) method="post" {
                select name="station" aria-label=(t!("select_station_to_add_to_route")) required {
                    @for station in &stations_not_in_route {
                        option value=(station.id) {
                            (station.name)
                        }
                    }
                    input type="submit" value=(t!("add"));
                }
        }
        }
        h2 { (t!("teams_with_this_route")) }
        @if teams.is_empty() {
            article {
                (t!("no_team_with_this_route"))
                a role="button" href="/admin/team" {
                    (t!("go_to_teams"))
                }

            }
        } @else {
            ol {
                @for team in &teams {
                    li {
                        a href=(format!("/admin/team/{}", team.id)) {
                            (team.name)
                        }
                    }
                }
            }
        }

    };
    Ok(page(content, session, false).await)
}

#[derive(Deserialize)]
struct UpdateNameForm {
    name: String,
}
async fn update_name(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
    Form(form): Form<UpdateNameForm>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, id).await else {
        er!(session, t!("route_update_err_nonexisting", id = id));

        return Redirect::to("/admin/route");
    };

    route.update_name(&db, &form.name).await;

    suc!(
        session,
        t!("route_new_name", old = route.name, new = form.name)
    );

    Redirect::to(&format!("/admin/route/{id}"))
}

#[derive(Deserialize)]
struct AddStationForm {
    station: i64,
}
async fn add_station(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
    Form(form): Form<AddStationForm>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, id).await else {
        er!(session, t!("route_update_err_nonexisting", id = id));

        return Redirect::to("/admin/route");
    };
    let Some(station) = Station::find_by_id(&db, form.station).await else {
        er!(
            session,
            t!(
                "route_cant_add_nonexisting_station",
                id = form.station,
                route = route.name
            )
        );

        return Redirect::to(&format!("/admin/route/{id}"));
    };

    match route.add_station(&db, &station).await {
        Ok(()) => suc!(
            session,
            t!(
                "station_added_to_route",
                station = station.name,
                route = route.name
            )
        ),
        Err(e) => er!(
            session,
            t!(
                "station_can_only_be_added_once_to_route",
                station = station.name,
                route = route.name,
                err = e
            )
        ),
    }

    Redirect::to(&format!("/admin/route/{id}"))
}

async fn delete_station(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path((route_id, station_id)): axum::extract::Path<(i64, i64)>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, route_id).await else {
        er!(
            session,
            t!("cant_remove_station_from_nonexisting_route", id = route_id)
        );

        return Redirect::to("/admin/route");
    };
    let Some(station) = Station::find_by_id(&db, station_id).await else {
        er!(
            session,
            t!(
                "cant_remove_nonexisting_station_from_route",
                id = station_id,
                route = route.name
            )
        );

        return Redirect::to(&format!("/admin/route/{route_id}"));
    };

    if route.delete_station(&db, &station).await {
        suc!(
            session,
            t!(
                "station_deleted_from_route",
                station = station.name,
                route = route.name
            )
        );
    } else {
        er!(
            session,
            t!(
                "cant_remove_station_from_route_if_not_included",
                station = station.name,
                route = route.name
            )
        );
    }

    Redirect::to(&format!("/admin/route/{route_id}"))
}

async fn move_station_higher(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path((route_id, station_id)): axum::extract::Path<(i64, i64)>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, route_id).await else {
        er!(session, t!("route_update_err_nonexisting", id = route_id));

        return Redirect::to("/admin/route");
    };
    let Some(station) = Station::find_by_id(&db, station_id).await else {
        er!(
            session,
            t!(
                "cant_move_nonexisting_station",
                id = station_id,
                route = route.name
            )
        );

        return Redirect::to(&format!("/admin/route/{route_id}"));
    };

    if route.move_station_higher(&db, &station).await {
        suc!(
            session,
            t!(
                "station_succ_moved_in_route",
                station = station.name,
                route = route.name
            )
        );
    } else {
        er!(
            session,
            t!(
                "cant_move_station_in_route_if_not_included",
                station = station.name,
                route = route.name
            )
        );
    }

    Redirect::to(&format!("/admin/route/{route_id}"))
}

pub(super) fn routes() -> Router<AppState> {
    Router::new()
        .route("/", get(index))
        .route("/", post(create))
        .route("/{id}/delete", get(delete))
        .route("/{id}", get(view))
        .route("/{id}/name", post(update_name))
        .route("/{id}/add-station", post(add_station))
        .route(
            "/{route_id}/delete-station/{station_id}",
            get(delete_station),
        )
        .route(
            "/{route_id}/move-station-higher/{station_id}",
            get(move_station_higher),
        )
}
