use crate::{auth::User, er, partials::page, suc, AppState};
use axum::{
    extract::State,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Form, Router,
};
use maud::{html, Markup, PreEscaped};
use serde::Deserialize;
use sqlx::SqlitePool;
use std::sync::Arc;
use tower_sessions::Session;

#[derive(Deserialize)]
struct CreateForm {
    name: String,
}

async fn create(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    Form(form): Form<CreateForm>,
) -> impl IntoResponse {
    let id = match User::create(&db, &form.name).await {
        Ok(id) => {
            suc!(session, t!("user_created", name = form.name));
            id
        }
        Err(e) => {
            er!(session, t!("user_create_error", name = form.name, err = e));

            return Redirect::to("/admin/user");
        }
    };

    Redirect::to(&format!("/admin/user/{id}"))
}

async fn delete(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> impl IntoResponse {
    let Some(user) = User::find_by_id(&db, id).await else {
        er!(session, t!("nonexisting_user", id = id));

        return Redirect::to("/admin/user");
    };

    match user.delete(&db).await {
        Ok(()) => suc!(session, t!("user_deleted", name = user.name)),
        Err(e) => er!(
            session,
            t!(
                "user_delete_error_already_in_use",
                name = user.name,
                err = e
            )
        ),
    }

    Redirect::to("/admin/user")
}

async fn view(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> Result<Markup, impl IntoResponse> {
    let Some(user) = User::find_by_id(&db, id).await else {
        er!(session, t!("nonexisting_user", id = id));
        return Err(Redirect::to("/admin/user"));
    };

    // maybe switch to maud-display impl of team
    let content = html! {
        h1 {
            a href="/admin/user" { "↩️" }
            (t!("admin"))
            " "
            (user.name)
        }
        article {
            details {
                summary {
                    (t!("edit_username"))
                    " ✏️"
                }
                form action=(format!("/admin/user/{}/name", user.id)) method="post" {
                    input type="text" name="name" value=(user.name) required;
                    input type="submit" value=(t!("save"));
                }

            }
        }
        table {
            tbody {
                tr {
                    th scope="row" { (t!("name")) };
                    td {
                        (user.name)
                    }
                }
                tr {
                    th scope="row" { (t!("pw")) };
                    td {
                        @if let Some(code) = user.require_new_password_code {
                            a href=(format!("/user/{}/set-pw/{}", user.id, code)){
                                (t!("login_link"))
                            }
                        } @else {
                            a href=(format!("/admin/user/{}/new-pw", user.id))
                                onclick=(format!("return confirm('{}');", t!("confirm_new_admin_link"))) {
                                    (t!("new_admin_link"))
                            }
                        }
                    }
                }
            }
        }
    };
    Ok(page(content, session, true).await)
}

#[derive(Deserialize)]
struct UpdateNameForm {
    name: String,
}
async fn update_name(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
    Form(form): Form<UpdateNameForm>,
) -> impl IntoResponse {
    let Some(user) = User::find_by_id(&db, id).await else {
        er!(session, t!("nonexisting_user", id = id));
        return Redirect::to("/admin/user");
    };

    user.update_name(&db, &form.name).await;

    suc!(
        session,
        t!("new_user_name", old = user.name, new = form.name)
    );

    Redirect::to(&format!("/admin/user/{id}"))
}

async fn new_pw(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> impl IntoResponse {
    let Some(user) = User::find_by_id(&db, id).await else {
        er!(session, t!("nonexisting_user", id = id));
        return Redirect::to("/admin/user");
    };

    user.new_pw(&db).await;

    suc!(session, t!("succ_new_admin_link", user = user.name));

    Redirect::to(&format!("/admin/user/{id}"))
}

async fn index(State(db): State<Arc<SqlitePool>>, session: Session) -> Markup {
    let users = User::all(&db).await;

    let content = html! {
        h1 {
            a href="/admin" { "↩️" }
            (t!("admins"))
        }
        article {
            em { "Admins" }
            " "
            (PreEscaped(t!("admin_expl_without_first_word")))
        }
        article {
            h2 { (t!("new_admin")) }
                form action="/admin/user" method="post" {
                        fieldset role="group" {
                            input type="text" name="name" placeholder=(t!("name")) required;
                            input type="submit" value=(t!("new_admin"));
                        }
                }
        }
        ul {
                @for user in &users {
                    li {
                         a href=(format!("/admin/user/{}", user.id)){
                            (user.name)
                         }
                         a href=(format!("/admin/user/{}/delete", user.id))
                           onclick=(format!("return confirm('{}');", t!("confirm_admin_delete"))) {
                             "🗑️"
                         }
                    }
                }
        }
    };
    page(content, session, false).await
}

pub(super) fn routes() -> Router<AppState> {
    Router::new()
        .route("/", get(index))
        .route("/", post(create))
        .route("/{id}", get(view))
        .route("/{id}/delete", get(delete))
        .route("/{id}/new-pw", get(new_pw))
        .route("/{id}/name", post(update_name))
}
